@@ -4,58 +4,56 @@
 
 from __future__ import absolute_import, division, print_function, unicode_literals
 
+import codecs
 import sys, os
 
-from os.path import abspath, basename, exists, isdir, isfile, join
+from os.path import isdir, isfile, join
 
 from conda.cli import common as cli_common
 from conda.cli.main import generate_parser, init_loggers
 from conda.base.context import context
-from conda.core.index import calculate_channel_urls, check_whitelist, _supplement_index_with_system #, get_index
-from conda.models.channel import Channel, prioritize_channels
-from conda.models.records import PackageRecord
 from conda.models.match_spec import MatchSpec
-from conda.cli.main_list import list_packages
 from conda.core.prefix_data import PrefixData
-from conda.misc import clone_env, explicit, touch_nonadmin
-from conda.common.serialize import json_dump
+from conda.misc import explicit, touch_nonadmin
 from conda.cli.common import specs_from_url, confirm_yn, check_non_admin, ensure_name_or_prefix
-from conda.core.subdir_data import SubdirData
+from conda.history import History
 from conda.core.link import UnlinkLinkTransaction, PrefixSetup
 from conda.cli.install import check_prefix, clone, print_activate
-from conda.base.constants import ChannelPriority, ROOT_ENV_NAME, UpdateModifier, DepsModifier
-from conda.core.solve import diff_for_unlink_link_precs
+from conda.base.constants import ChannelPriority, UpdateModifier, DepsModifier
+from conda.core.solve import get_pinned_specs
 from conda.core.envs_manager import unregister_env
 from conda.core.package_cache_data import PackageCacheData
+from conda.common.compat import on_win
 
 # create support
 from conda.common.path import paths_equal
-from conda.exceptions import (CondaExitZero, CondaImportError, CondaOSError, CondaSystemExit,
+from conda.exceptions import (CondaExitZero, CondaOSError, CondaSystemExit,
                               CondaValueError, DirectoryNotACondaEnvironmentError, CondaEnvironmentError,
-                              DirectoryNotFoundError, DryRunExit, EnvironmentLocationNotFound,
+                              DryRunExit, EnvironmentLocationNotFound,
                               NoBaseEnvironmentError, PackageNotInstalledError, PackagesNotFoundError,
-                              TooManyArgumentsError, UnsatisfiableError)
+                              TooManyArgumentsError)
 
 from conda.gateways.disk.create import mkdir_p
 from conda.gateways.disk.delete import rm_rf, delete_trash, path_is_clean
 from conda.gateways.disk.test import is_conda_environment
 
-from conda._vendor.boltons.setutils import IndexedSet
-from conda.models.prefix_graph import PrefixGraph
-
 from logging import getLogger
 
-import json
-import tempfile
-import logging
-
 import mamba
 import mamba.mamba_api as api
 
-from mamba.utils import get_index, to_package_record_from_subjson, init_api_context
+from mamba.utils import get_index, to_package_record_from_subjson, init_api_context, get_installed_jsonfile, to_txn
+
+if sys.version_info < (3, 2):
+    sys.stdout = codecs.lookup('utf-8')[-1](sys.stdout)
+elif sys.version_info < (3, 7):
+    sys.stdout = codecs.getwriter("utf-8")(sys.stdout.detach())
+else:
+    sys.stdout.reconfigure(encoding='utf-8')
+
 
 log = getLogger(__name__)
-stderrlog = getLogger('mamba.stderr')
+stderrlog = getLogger('conda.stderr')
 
 banner = f"""
                   __    __    __    __
@@ -86,28 +84,6 @@ class MambaException(Exception):
 
 solver_options = [(api.SOLVER_FLAG_ALLOW_DOWNGRADE, 1)]
 
-def get_installed_packages(prefix, show_channel_urls=None):
-    result = {'packages': {}}
-
-    # Currently, we need to have pip interop disabled :/
-    installed = {rec: rec for rec in PrefixData(prefix, pip_interop_enabled=False).iter_records()}
-
-    # add virtual packages as installed packages
-    # they are packages installed on the system that conda can do nothing about (e.g. glibc)
-    # if another version is needed, installation just fails
-    # they don't exist anywhere (they start with __)
-    _supplement_index_with_system(installed)
-    installed = list(installed)
-
-    for prec in installed:
-        json_rec = prec.dist_fields_dump()
-        json_rec['depends'] = prec.depends
-        json_rec['build'] = prec.build
-        result['packages'][prec.fn] = json_rec
-
-    return installed, result
-
-
 def specs_from_args(args, json=False):
 
     def arg2spec(arg, json=False, update=False):
@@ -128,49 +104,6 @@ def specs_from_args(args, json=False):
 
     return [arg2spec(arg, json=json) for arg in args]
 
-
-def to_txn(specs_to_add, specs_to_remove, prefix, to_link, to_unlink, index=[]):
-    to_link_records, to_unlink_records = [], []
-
-    prefix_data = PrefixData(prefix)
-    final_precs = IndexedSet(prefix_data.iter_records())
-
-    lookup_dict = {}
-    for _, c in index:
-        lookup_dict[str(c)] = c
-
-    for c, pkg in to_unlink:
-        for i_rec in installed_pkg_recs:
-            if i_rec.fn == pkg:
-                final_precs.remove(i_rec)
-                to_unlink_records.append(i_rec)
-                break
-        else:
-            print("No package record found!")
-
-    for c, pkg, jsn_s in to_link:
-        sdir = lookup_dict[c]
-        rec = to_package_record_from_subjson(sdir, pkg, jsn_s)
-        final_precs.add(rec)
-        to_link_records.append(rec)
-
-    unlink_precs, link_precs = diff_for_unlink_link_precs(prefix,
-                                                          final_precs=IndexedSet(PrefixGraph(final_precs).graph),
-                                                          specs_to_add=specs_to_add,
-                                                          force_reinstall=context.force_reinstall)
-
-    pref_setup = PrefixSetup(
-        target_prefix  = prefix,
-        unlink_precs   = unlink_precs,
-        link_precs     = link_precs,
-        remove_specs   = specs_to_remove,
-        update_specs   = specs_to_add,
-        neutered_specs = ()
-    )
-
-    conda_transaction = UnlinkLinkTransaction(pref_setup)
-    return conda_transaction
-
 use_mamba_experimental = False
 
 def handle_txn(unlink_link_transaction, prefix, args, newenv, remove_op=False):
@@ -207,17 +140,6 @@ def handle_txn(unlink_link_transaction, prefix, args, newenv, remove_op=False):
         actions = unlink_link_transaction._make_legacy_action_groups()[0]
         cli_common.stdout_json_success(prefix=prefix, actions=actions)
 
-installed_pkg_recs = None
-
-def get_installed_jsonfile(prefix):
-    global installed_pkg_recs
-    installed_pkg_recs, output = get_installed_packages(prefix, show_channel_urls=True)
-    installed_json_f = tempfile.NamedTemporaryFile('w', delete=False)
-    installed_json_f.write(json_dump(output))
-    installed_json_f.flush()
-    return installed_json_f
-
-
 def remove(args, parser):
     if not (args.all or args.package_names):
         raise CondaValueError('no package names supplied,\n'
@@ -269,7 +191,7 @@ def remove(args, parser):
         if not context.quiet:
             print("Removing specs: {}".format([s.conda_build_form() for s in specs]))
 
-        installed_json_f = get_installed_jsonfile(prefix)
+        installed_json_f, installed_pkg_recs = get_installed_jsonfile(prefix)
 
         mamba_solve_specs = [s.conda_build_form() for s in specs]
 
@@ -309,7 +231,7 @@ def remove(args, parser):
         specs_to_add = [MatchSpec(m) for m in mmb_specs[0]]
         specs_to_remove = [MatchSpec(m) for m in mmb_specs[1]]
 
-        conda_transaction = to_txn(specs_to_add, specs_to_remove, prefix, to_link, to_unlink)
+        conda_transaction = to_txn(specs_to_add, specs_to_remove, prefix, to_link, to_unlink, installed_pkg_recs)
         handle_txn(conda_transaction, prefix, args, False, True)
 
 def install(args, parser, command='install'):
@@ -328,6 +250,7 @@ def install(args, parser, command='install'):
     isupdate = bool(command == 'update')
     if isupdate:
         solver_task = api.SOLVER_UPDATE
+        solver_options.clear()
 
     if newenv:
         ensure_name_or_prefix(args, command)
@@ -346,7 +269,9 @@ def install(args, parser, command='install'):
 
     if not newenv:
         if isdir(prefix):
-            delete_trash(prefix)
+            if on_win:
+                delete_trash(prefix)
+
             if not isfile(join(prefix, 'conda-meta', 'history')):
                 if paths_equal(prefix, context.conda_prefix):
                     raise NoBaseEnvironmentError()
@@ -357,7 +282,7 @@ def install(args, parser, command='install'):
                 # fall-through expected under normal operation
                 pass
         else:
-            if args.mkdir:
+            if hasattr(args, "mkdir") and args.mkdir:
                 try:
                     mkdir_p(prefix)
                 except EnvironmentError as e:
@@ -392,6 +317,14 @@ def install(args, parser, command='install'):
 
     specs = []
 
+    index_args = {
+        'use_cache': args.use_index_cache,
+        'channel_urls': context.channels,
+        'unknown': args.unknown,
+        'prepend': not args.override_channels,
+        'use_local': args.use_local
+    }
+
     if args.file:
         file_specs = []
         for fpath in args.file:
@@ -407,6 +340,7 @@ def install(args, parser, command='install'):
 
     specs.extend(specs_from_args(args_packages, json=context.json))
 
+    # update channels from package specs (e.g. mychannel::mypackage adds mychannel)
     channels = [c for c in context.channels]
     for spec in specs:
         # CONDA TODO: correct handling for subdir isn't yet done
@@ -414,13 +348,7 @@ def install(args, parser, command='install'):
         if spec_channel and spec_channel not in channels:
             channels.append(spec_channel)
 
-    index_args = {
-        'use_cache': args.use_index_cache,
-        'channel_urls': channels,
-        'unknown': args.unknown,
-        'prepend': not args.override_channels,
-        'use_local': args.use_local
-    }
+    index_args['channel_urls'] = channels
 
     index = get_index(channel_urls=index_args['channel_urls'],
                       prepend=index_args['prepend'], platform=None,
@@ -461,10 +389,7 @@ def install(args, parser, command='install'):
 
         channel_json.append((chan, subdir, priority, subpriority))
 
-    # for c in channel_json:
-    #     print(c[0], c[2], c[3])
-
-    installed_json_f = get_installed_jsonfile(prefix)
+    installed_json_f, installed_pkg_recs = get_installed_jsonfile(prefix)
 
     if isinstall and args.revision:
         get_revision(args.revision, json=context.json)
@@ -472,23 +397,49 @@ def install(args, parser, command='install'):
         raise CondaValueError("too few arguments, "
                               "must supply command line package specs or --file")
 
+    installed_names = [i_rec.name for i_rec in installed_pkg_recs]
     # for 'conda update', make sure the requested specs actually exist in the prefix
     # and that they are name-only specs
     if isupdate and context.update_modifier == UpdateModifier.UPDATE_ALL:
-        # Note: History(prefix).get_requested_specs_map()
-        print("Currently, mamba can only update explicit packages! (e.g. mamba update numpy python ...)")
-        exit()
+        history_dict = History(prefix).get_requested_specs_map()
+        pins = {
+            pin.name: pin
+            for pin in get_pinned_specs(prefix)
+        }
+        # for key, match_spec in history_dict.items():
+        for key in installed_names:
+            if key == 'python':
+                i = installed_names.index('python')
+                version = installed_pkg_recs[i].version
+                py_ver = ".".join(version.split(".")[:2]) + '.*'
+                # specs.append(MatchSpec(name="python", version=py_ver))
+            else:
+                if key in pins:
+                    specs.append(pins[key])
+                else:
+                    specs.append(MatchSpec(key))
 
-    if isupdate and context.update_modifier != UpdateModifier.UPDATE_ALL:
         prefix_data = PrefixData(prefix)
         for s in args_packages:
             s = MatchSpec(s)
             if not s.is_name_only_spec:
                 raise CondaValueError("Invalid spec for 'conda update': %s\n"
-                                 "Use 'conda install' instead." % s)
+                                      "Use 'conda install' instead." % s)
             if not prefix_data.get(s.name, None):
                 raise PackageNotInstalledError(prefix, s.name)
 
+    elif context.update_modifier == UpdateModifier.UPDATE_DEPS:
+        # find the deps for each package and add to the update job
+        # solver_task |= api.SOLVER_FORCEBEST
+        final_specs = specs
+        for spec in specs:
+            prec = installed_pkg_recs[installed_names.index(spec.name)]
+            for dep in prec.depends:
+                ms = MatchSpec(dep)
+                if ms.name != 'python':
+                    final_specs.append(MatchSpec(ms.name))
+        specs = set(final_specs)
+
     if newenv and args.clone:
         if args.packages:
             raise TooManyArgumentsError(0, len(args.packages), list(args.packages),
@@ -499,16 +450,18 @@ def install(args, parser, command='install'):
         print_activate(args.name if args.name else prefix)
         return
 
-    spec_names = [s.name for s in specs]
 
     if not (context.quiet or context.json):
-        print("\nLooking for: {}\n".format(spec_names))
+        print("\nLooking for: {}\n".format([str(s) for s in specs]))
+
+    spec_names = [s.name for s in specs]
 
     # If python was not specified, check if it is installed.
     # If yes, add the installed python to the specs to prevent updating it.
     python_constraint = None
+    additional_specs = []
+
     if 'python' not in spec_names:
-        installed_names = [i_rec.name for i_rec in installed_pkg_recs]
         if 'python' in installed_names:
             i = installed_names.index('python')
             version = installed_pkg_recs[i].version
@@ -520,10 +473,12 @@ def install(args, parser, command='install'):
 
     repos = []
 
-    # add installed
-    if use_mamba_experimental:
+    if use_mamba_experimental or context.force_reinstall:
         prefix_data = api.PrefixData(context.target_prefix)
         prefix_data.load()
+
+    # add installed
+    if use_mamba_experimental:
         repo = api.Repo(pool, prefix_data)
         repos.append(repo)
     else:
@@ -536,16 +491,26 @@ def install(args, parser, command='install'):
         repo.set_priority(priority, subpriority)
         repos.append(repo)
 
-    solver = api.Solver(pool, solver_options)
+    if context.force_reinstall:
+        solver = api.Solver(pool, solver_options, prefix_data)
+    else:
+        solver = api.Solver(pool, solver_options)
+
+    solver.set_postsolve_flags(
+        [(api.MAMBA_NO_DEPS, context.deps_modifier == DepsModifier.NO_DEPS), 
+         (api.MAMBA_ONLY_DEPS, context.deps_modifier == DepsModifier.ONLY_DEPS),
+         (api.MAMBA_FORCE_REINSTALL, context.force_reinstall)]
+    )
     solver.add_jobs(mamba_solve_specs, solver_task)
 
+    # as a security feature this will _always_ attempt to upgrade certain packages
+    for a_pkg in [_.name for _ in context.aggressive_update_packages]:
+        if a_pkg in installed_names:
+            solver.add_jobs([a_pkg], api.SOLVER_UPDATE)
+
     if python_constraint:
         solver.add_constraint(python_constraint)
 
-    solver.set_postsolve_flags(
-        [(api.MAMBA_NO_DEPS, context.deps_modifier == DepsModifier.NO_DEPS), 
-         (api.MAMBA_ONLY_DEPS, context.deps_modifier == DepsModifier.ONLY_DEPS)]
-    )
     success = solver.solve()
     if not success:
         print(solver.problems_to_str())
@@ -572,7 +537,7 @@ def install(args, parser, command='install'):
 
         transaction.execute(prefix_data, PackageCacheData.first_writable().pkgs_dir)
     else:
-        conda_transaction = to_txn(specs_to_add, specs_to_remove, prefix, to_link, to_unlink, index)
+        conda_transaction = to_txn(specs_to_add, specs_to_remove, prefix, to_link, to_unlink, installed_pkg_recs, index)
         handle_txn(conda_transaction, prefix, args, newenv)
 
     try:
@@ -611,6 +576,15 @@ def update(args, parser):
     install(args, parser, 'update')
 
 def repoquery(args, parser):
+    if not args.subcmd:
+        print("repoquery needs a subcommand (search, depends or whoneeds)")
+        print("eg:")
+        print("    $ mamba repoquery search xtensor\n")
+        exit(1)
+
+    if args.platform:
+        context._subdirs = (args.platform, 'noarch')
+
     prefix = context.target_prefix
 
     init_api_context()
@@ -623,22 +597,29 @@ def repoquery(args, parser):
         'use_local': args.use_local
     }
 
-    installed_json_f = get_installed_jsonfile(prefix)
+    installed_json_f, installed_pkg_recs = get_installed_jsonfile(prefix)
 
     pool = api.Pool()
     repos = []
 
-    # add installed
-    repo = api.Repo(pool, "installed", installed_json_f.name, "")
-    repo.set_installed()
-    repos.append(repo)
-
     only_installed = True
     if args.subcmd == "search" and args.installed == False:
         only_installed = False
-    elif args.all_channels:
+    elif args.all_channels or len(args.channel):
         only_installed = False
 
+    if only_installed and args.no_installed:
+        print("No channels selected.")
+        print("Activate -a to search all channels.")
+        exit(1)
+
+    if not args.no_installed:
+        # add installed
+        repo = api.Repo(pool, "installed", installed_json_f.name, "")
+        repo.set_installed()
+        repos.append(repo)
+
+
     if not only_installed:
         index = get_index(channel_urls=index_args['channel_urls'],
                       prepend=index_args['prepend'], platform=None,
@@ -721,6 +702,9 @@ Examples:
         help=SUPPRESS
     )
 
+    package_cmds.add_argument('-p', '--platform')
+    package_cmds.add_argument('--no-installed', action='store_true')
+
     package_cmds.add_argument(
         "-a", "--all-channels",
         action="store_true",
@@ -733,26 +717,27 @@ Examples:
         action="store_true"
     )
 
-    subsub_parser.add_parser('whoneeds',
+    c1 = subsub_parser.add_parser('whoneeds',
         help='shows packages that depends on this package',
         parents=[package_cmds, view_cmds]
     )
 
-    subsub_parser.add_parser('depends',
+    c2 = subsub_parser.add_parser('depends',
         help='shows packages that depends on this package',
         parents=[package_cmds, view_cmds]
     )
 
-    subsub_parser.add_parser('search',
+    c3 = subsub_parser.add_parser('search',
         help='shows packages that depends on this package',
         parents=[package_cmds]
     )
 
     from conda.cli import conda_argparse
-    conda_argparse.add_parser_channels(p)
-    conda_argparse.add_parser_networking(p)
-    conda_argparse.add_parser_known(p)
-    conda_argparse.add_parser_json(p)
+    for cmd in (c1, c2, c3):
+        conda_argparse.add_parser_channels(cmd)
+        conda_argparse.add_parser_networking(cmd)
+        conda_argparse.add_parser_known(cmd)
+        conda_argparse.add_parser_json(cmd)
 
     p.set_defaults(func='.main_repoquery.execute')
     return p
@@ -807,7 +792,7 @@ def main(*args, **kwargs):
 
     args = tuple(ensure_text_type(s) for s in args)
 
-    if len(args) > 2 and args[1] == 'env' and args[2] == 'create':
+    if len(args) > 2 and args[1] == 'env' and args[2] in ('create', 'update'):
         # special handling for conda env create!
         from mamba import mamba_env
         return mamba_env.main()
